/*
	MOD plugin for SoundPlay.
	Copyright 1999 Marco Nelissen (marcone@xs4all.nl)
	libMikMod player copyright Miodrag Vallat and others.

	This is an example of a data producing plugin for SoundPlay.
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <AppKit.h>
#include <InterfaceKit.h>
#include <MediaDefs.h>
#include <Path.h>
#include "mikmod.h"
#include "SPModPlugin.h"


/* ===========================================================
	Plugin-descriptor structure.
	This structure is used to gather information about a plugin
	and	instantiate and destroy specific instances of a plugin.
*/
static plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"sp-mod",1, PLUGIN_IS_DECODER|PLUGIN_IS_SINGLE_CONTEXT,

	"MOD player",
	"By Marco Nelissen.\n\n"
	"Based on libMikMod by Jean-Paul Mikkers and others.\n",
	&aboutplugin,
	NULL, // configure
	&getplugin,
	&destroyplugin
};

/* ===========================================================
	Plugin-operations structure.
	This is used to actually perform operations on a plugin.
	It is returned by the getplugin() function from the 
	plugin_descriptor structure above.
*/
static decoder_plugin_ops ops={
	PLUGIN_DECODER_MAGIC,PLUGIN_DECODER_VERSION,

	NULL,	// open
	NULL,	// close
	info,
	readdata,
	NULL,	// play
	NULL,	// stop
	NULL,	// setvolume
	NULL,	// setpitchandspeed
	seek,
	position,
	NULL  // bufferamount
};


/* ===========================================================
	Return the list of plugins in this file.
*/
static plugin_descriptor *plugs[]={
	&plugin,
	0
};

plugin_descriptor **get_plugin_list(void)
{
	return plugs;
}

/* ===========================================================
	Return the list of filetypes supported by all the plugins
	in this file
*/
static supported_type supportedtypes[]=
{
	{"audio/x-xm",		"1-32 channels FastTracker II module",							"Fasttracker II module",	"xm", NULL },
	{"audio/x-s3m",		"1-16 channels ScreamTracker 3.x module",						"ScreamTracker 3.x module",	"s3m",NULL },
	{"audio/x-mod",		"ProTracker and various module formats with 4 or more channels","MOD sound module",			"mod",NULL },
	{"audio/x-uni",		"UniTracker module",											"Unitracker Module",		"uni",NULL },
	{"audio/x-ult",		"1-32 channels UltraTracker module",							"UltraTracker module",		"ult",NULL },
	{"audio/x-stx",		"4 channels STMIK module",										"STMIK Module",				"stx",NULL },
	{"audio/x-stm",		"4 channels ScreamTracker 2.x module",							"ScreamTracker 2.x module",	"stm", NULL},
	{"audio/x-mtm",		"2-32 channels MultiTracker module format",						"MultiTracker module",		"mtm", NULL },
	{"audio/x-med",		"MED module",													"MED module",				"med", NULL },
	{"audio/x-m15",		"M15 module",													"MOD: m15",					"m15", NULL },
	{"audio/x-it",		"1-64 channels ImpulseTracker module",							"Impulse Tracker module",	"it", NULL },
	{"audio/x-imf",		"1-32 channels Imago Orpheus module",							"Imago Orpheus module",		"imf", NULL },
	{"audio/x-gdm",		"1-32 channels General DigiMusic module",						"General DigiMusic module",	"gdm", NULL },
	{"audio/x-far",		"16 channels Farandole module",									"Farandole module",			"far", NULL },
	{"audio/x-amf",		"1-32 channels DSMI module",									"DSMI module",				"amf", NULL },
	{"audio/x-669",		"8 channels 669 Tracker module",								"669 Tracker module",		"669", NULL },
	{NULL,				NULL,															NULL,						NULL,NULL}
};

supported_type *get_supported_types(void)
{
	return supportedtypes;
}


/* ===========================================================
	Implementation of the plugin_descriptor functions
*/
static bool mmRegistered=false;
static void *getplugin(void **data, const char *name, const char *header, uint32 size, plugin_info *info, SoundPlayBufferedReader *reader)
{
	if(!mmRegistered)
	{
		mmRegistered=true;
		MikMod_RegisterAllLoaders();
		MikMod_RegisterAllDrivers();
		MikMod_Init();
		md_mode|=DMODE_SOFT_MUSIC;
	}

	MODULE *mf = Player_Load((char*)name, 128, 0 );
	if(mf)
	{
		Player_Start(mf);
		*data=(void*)mf;
		return &ops;
	}

	Player_Free(mf);
	MikMod_Exit();
	mmRegistered=false;
	return NULL;
}

static void destroyplugin(void *ops,void *_mf)
{
	MODULE *mf = (MODULE*)_mf;
	Player_Stop();
	Player_Free(mf);
	if(mmRegistered)
	{
		MikMod_Exit();
		mmRegistered=false;
	}
}

// Normally you wouldn't implement the about function for such a simple
// message. It is done here for completeness.
static void aboutplugin()
{
	BAlert *alert=new BAlert("About MOD player",
		"MikMod MOD player.\n"
		"By Marco Nelissen.\n\n"
		"Based on libMikMod by Jean-Paul Mikkers and others.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go(NULL);
}


/* ===========================================================
	Implementation of the decoder_plugin_ops functions
*/
int32 readdata(void *_mf, char *buf, ulong count)
{
//	MODULE *mf = (MODULE*)_mf;
	int32 Cnt = VC_WriteBytes((SBYTE*)buf,count*4);
	return Cnt/4;
}

status_t info(void *_mf, file_info *info)
{
	MODULE *mf = (MODULE*)_mf;

// Fill in the file_info structure
	strcpy(info->name,"");
	strcpy(info->typedesc,mf->modtype);
	strcpy(info->mimetype,"audio/x-mod"); // I should check the module type, but I'm lazy...
	info->samplerate=44100;
	info->bitrate=44100*4*8;
	info->numchannels=2;
	info->granularity=1;
	info->framecount=1000000;
	info->samplesize=2;
	if(B_HOST_IS_LENDIAN)
		info->byteorder=B_LITTLE_ENDIAN;
	else
		info->byteorder=B_BIG_ENDIAN;
	info->sampleformat=B_LINEAR_SAMPLES;
	info->flags=PLUGIN_FILELENGTH_UNKNOWN|PLUGIN_NO_ELAPSEDTIME|PLUGIN_POSITION_IS_RELATIVE;
	
	return B_OK;
}

static status_t	seek(void *_mf, uint32 pos)
{
	MODULE *mf = (MODULE*)_mf;
	Player_SetPosition( (mf->numpos*pos)/1000000 );
	return B_OK;
}

static uint32 position(void *_mf)
{
	MODULE *mf = (MODULE*)_mf;
	uint32 pos;
	pos=(mf->sngpos*1000000)/mf->numpos+(mf->patpos*(1000000/mf->numpos))/mf->numrow;
	return pos;
}

int main()
{
	BApplication app("application/x-marcone-modplugin");
	BAlert *alert=new BAlert("About MOD player",
		"MOD plugin for SoundPlay.\n"
		"By Marco Nelissen.\n\n"
		"Based on libMikMod by Jean-Paul Mikkers and others.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
}

